################################################################################
# Copyright (c) 2017-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpservices.interfaces`."""

import unittest
from unittest import mock

import netifaces

from katsdpservices import get_interface_address


class TestGetInterfaceAddress(unittest.TestCase):
    def test_none(self):
        """Passing None returns None"""
        self.assertIsNone(get_interface_address(None))

    def test_invalid(self):
        """Passing a non-existent interface raises :exc:`ValueError`"""
        with self.assertRaises(ValueError):
            get_interface_address('not_an_interface_name')

    def test_no_ipv4(self):
        """An interface with no IPv4 entries raises :exc:`ValueError`"""
        with mock.patch('netifaces.ifaddresses', return_value={
            netifaces.AF_LINK: [{'addr': 'de:ad:be:ef:ca:fe', 'broadcast': 'ff:ff:ff:ff:ff:ff'}]
        }):
            with self.assertRaises(ValueError):
                get_interface_address('wlan1')

    def test_valid(self):
        with mock.patch('netifaces.ifaddresses', return_value={
            netifaces.AF_LINK: [{'addr': 'de:ad:be:ef:ca:fe', 'broadcast': 'ff:ff:ff:ff:ff:ff'}],
            netifaces.AF_INET: [{'addr': '192.168.1.1', 'broadcast': '192.168.1.255',
                                 'netmask': '255.255.255.0'}]
        }) as m:
            self.assertEqual('192.168.1.1', get_interface_address('eth1'))
            m.assert_called_with('eth1')



################################################################################
# Copyright (c) 2017-2020, 2022, 2024 National Research Foundation (SARAO)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpservices.aiomonitor`."""

import asyncio
import unittest
from unittest import mock

import aiomonitor

from katsdpservices import ArgumentParser, start_aiomonitor, add_aiomonitor_arguments


class TestStartAiomonitor(unittest.TestCase):
    def setUp(self):
        self.parser = ArgumentParser()
        add_aiomonitor_arguments(self.parser)
        patcher = mock.patch('aiomonitor.start_monitor', autospec=True)
        self.mock_start = patcher.start()
        self.addCleanup(patcher.stop)
        self.loop = asyncio.new_event_loop()

    def test_no_aiomonitor(self):
        args = self.parser.parse_args([])
        with start_aiomonitor(self.loop, args, locals()):
            pass
        self.mock_start.assert_not_called()

    def test_defaults(self):
        args = self.parser.parse_args(['--aiomonitor'])
        locals_ = {'hello': 'world'}
        with start_aiomonitor(self.loop, args, locals_):
            pass
        self.mock_start.assert_called_once_with(
            loop=self.loop, host=aiomonitor.MONITOR_HOST,
            port=aiomonitor.MONITOR_PORT,
            console_port=aiomonitor.CONSOLE_PORT,
            webui_port=aiomonitor.MONITOR_WEBUI_PORT,
            locals=locals_)

    def test_explicit(self):
        args = self.parser.parse_args(
            ['--aiomonitor',
             '--aiomonitor-host', 'example.com',
             '--aiomonitor-port', '1234',
             '--aiomonitor-webui-port', '4321',
             '--aioconsole-port', '2345'])
        locals_ = {'hello': 'world'}
        with start_aiomonitor(self.loop, args, locals_):
            pass
        self.mock_start.assert_called_once_with(
            loop=self.loop, host='example.com',
            port=1234,
            console_port=2345,
            webui_port=4321,
            locals=locals_)



################################################################################
# Copyright (c) 2017-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpservices.restart`"""

import os
import signal
import time
import unittest
from unittest import mock

import katsdpservices


class TestRestart(unittest.TestCase):
    def _create_patch(self, *args, **kwargs):
        patcher = mock.patch(*args, **kwargs)
        mock_obj = patcher.start()
        self.addCleanup(patcher.stop)
        return mock_obj

    def setUp(self):
        self.execlp = self._create_patch('os.execlp', spec=True)
        self.addCleanup(signal.signal, signal.SIGHUP, signal.SIG_DFL)

    def test_restart(self):
        katsdpservices.restart_process()
        # assert_called_once_with can't specify an unknown number of arguments
        self.assertEqual(1, len(self.execlp.mock_calls))

    def test_restart_signal(self):
        katsdpservices.setup_restart()
        os.kill(os.getpid(), signal.SIGHUP)
        # Give time for asynchronous handling
        time.sleep(0.01)
        self.assertEqual(1, len(self.execlp.mock_calls))

    def test_restart_signal_with_false_callback(self):
        callback = mock.MagicMock()
        callback.return_value = False
        katsdpservices.setup_restart(restart_callback=callback)
        os.kill(os.getpid(), signal.SIGHUP)
        # Give time for asynchronous handling
        time.sleep(0.01)
        callback.assert_called_once_with()
        self.assertEqual(1, len(self.execlp.mock_calls))

    def test_restart_signal_with_true_callback(self):
        callback = mock.MagicMock()
        callback.return_value = True
        katsdpservices.setup_restart(restart_callback=callback)
        os.kill(os.getpid(), signal.SIGHUP)
        # Give time for asynchronous handling
        time.sleep(0.01)
        callback.assert_called_once_with()
        self.execlp.assert_not_called()



################################################################################
# Copyright (c) 2017-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Argument parser that can load defaults from a telescope state.

See :class:`ArgumentParser` for details.
"""

import argparse
try:
    import katsdptelstate.endpoint
except ImportError:
    pass


class _HelpAction(argparse.Action):
    """Class modelled on argparse._HelpAction that prints help for the
    main parser."""
    def __init__(self,
                 option_strings,
                 parser,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help=None):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)
        self._parser = parser

    def __call__(self, parser, namespace, values, option_string=None):
        self._parser.print_help()
        self._parser.exit()


class ArgumentParser(argparse.ArgumentParser):
    """Argument parser that can load defaults from a telescope state. It can be
    used as a drop-in replacement for `argparse.ArgumentParser`. It adds the
    options `--telstate` and `--name`. The first takes the hostname of a
    telescope state repository (optionally with a port). If specified,
    `parse_args` will first connect to this host and fetch defaults (which
    override the defaults specified by `add_argument`). The telescope state
    will also be available in the returned `argparse.Namespace`.

    If `name` is specified, it consists of a dot-separated list of
    identifiers, specifying a path through a tree of dictionaries of config.
    For example, `foo.0` will cause configuration to be searched in
    `config`, `config["foo"]`, `config["foo"]["0"]`, `config.foo` and
    `config.foo.0`. If configuration is found in multiple places, the most
    specific location will be used first, breaking ties to use `config.*`
    in preference to embedded dictionaries within `config`. It is not an error
    for one of these dictionaries not to exist, but it is an error if a name is
    found but is not a dictionary.

    New code generating config dictionaries is advised to use the `config.*`
    form, as it is more scalable (no need to fetch unrelated config).
    Sub-config embedded within `config` is supported for backwards
    compatibility.

    A side-effect of the implementation is that calling `parse_args` or
    `parse_known_args` permanently changes the defaults. A parser should thus
    only be used once and then thrown away. Also, because it changes the
    defaults rather than injecting actual arguments, argparse features like
    required arguments and mutually exclusive groups might not work as
    expected.

    Parameters
    ----------
    config_key : str, optional
        Name of the config dictionary within the telescope state (default: `config`)
    """

    _SPECIAL_NAMES = ['telstate', 'name']

    def __init__(self, *args, **kwargs):
        self.config_key = kwargs.pop('config_key', 'config')
        super().__init__(*args, **kwargs)
        # Create a separate parser that will extract only the special args
        self.config_parser = argparse.ArgumentParser(add_help=False)
        self.config_parser.add_argument('-h', '--help', action=_HelpAction,
                                        default=argparse.SUPPRESS, parser=self)
        for parser in [super(), self.config_parser]:
            parser.add_argument(
                '--telstate', help='Telescope state repository from which to retrieve config',
                metavar='HOST[:PORT]')
            parser.add_argument(
                '--name', type=str, default='',
                help='Name of this process for telescope state configuration')

    def _valid_keys(self):
        """Get set of keys that match defined arguments"""
        # _actions is a private member of the base class! But trying to avoid
        # accessing this is quite difficult if one wishes to handle cases like
        # :meth:`add_argument_group` and :meth:`add_mutually_exclusive_group`,
        # parent parsers etc.
        bad_names = [None, argparse.SUPPRESS, 'help'] + self._SPECIAL_NAMES
        return {action.dest for action in self._actions if action.dest not in bad_names}

    def _load_defaults(self, telstate, name):
        config_dict = telstate.get(self.config_key, {})
        parts = name.split('.')
        cur = config_dict
        config = cur.copy()
        split_name = self.config_key
        for part in parts:
            if cur is not None:
                cur = cur.get(part)
                if cur is not None:
                    config.update(cur)
            split_name += '.'
            split_name += part
            config.update(telstate.get(split_name, {}))

        valid_keys = self._valid_keys()
        defaults = {key: value for key, value in config.items() if key in valid_keys}
        super().set_defaults(**defaults)

    def set_defaults(self, **kwargs):
        for special in self._SPECIAL_NAMES:
            if special in kwargs:
                self.config_parser.set_defaults(**{special: kwargs.pop(special)})
        super().set_defaults(**kwargs)

    def parse_known_args(self, args=None, namespace=None):
        if namespace is None:
            namespace = argparse.Namespace()
        try:
            config_args, other = self.config_parser.parse_known_args(args)
        except argparse.ArgumentError:
            other = args
        else:
            if config_args.telstate is not None:
                try:
                    namespace.telstate_endpoint = \
                        katsdptelstate.endpoint.endpoint_parser(6379)(config_args.telstate)
                    namespace.telstate = katsdptelstate.TelescopeState(config_args.telstate)
                except katsdptelstate.ConnectionError as e:
                    self.error(str(e))
                namespace.name = config_args.name
                self._load_defaults(namespace.telstate, namespace.name)
            else:
                namespace.telstate_endpoint = None
        return super().parse_known_args(other, namespace)

    def add_aiomonitor_arguments(self):
        """Add a set of arguments for controlling aiomonitor.

        .. deprecated:: June 2019

            Use the :func:`~.add_aiomonitor_arguments` free function instead.
        """
        from .aiomonitor import add_aiomonitor_arguments
        add_aiomonitor_arguments(self)



################################################################################
# Copyright (c) 2017-2020, 2024, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Utilities to simplify starting aiomonitor"""

import inspect


class _DummyContext:
    """Context manager that does nothing"""
    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


def start_aiomonitor(loop, args, locals):
    """Optionally create and start aiomonitor, depending on command-line arguments.

    The return value should be used as a context manager e.g.::

        with start_aiomonitor(loop, args, locals=locals()):
            loop.run_until_complete(main())

    If ``--aiomonitor`` is not passed on the command line, it does not start
    aiomonitor, although the package must still be present.

    Parameters
    ----------
    loop : :class:`asyncio.AbstractEventLoop`
        Event loop to pass to the server
    args : :class:`argparse.Namespace`
        Command-line arguments, returned from a :class:`katsdpservices.ArgumentParser`
        created with ``aiomonitor=True``.
    locals : dict, optional
        Local variables, made available in aioconsole
    """
    import aiomonitor

    if not args.aiomonitor:
        return _DummyContext()
    else:
        # Add this only if aiomonitor is new enough to support it
        kwargs = {}
        if 'webui_port' in inspect.signature(aiomonitor.start_monitor).parameters:
            kwargs['webui_port'] = args.aiomonitor_webui_port
        return aiomonitor.start_monitor(
            loop=loop,
            host=args.aiomonitor_host,
            port=args.aiomonitor_port,
            console_port=args.aioconsole_port,
            locals=locals,
            **kwargs)


def add_aiomonitor_arguments(parser):
    """Add a set of arguments for controlling aiomonitor.

    See :func:`.start_aiomonitor` for details.

    The :option:`--aiomonitor-webui-port` option is added unconditionally,
    but ignored if aiomonitor is older than 0.6.0.

    Parameters
    ----------
    parser : :class:`argparse.ArgumentParser`
        Parser to which arguments will be added.
    """
    import aiomonitor

    # Fallback to providing our own copy of the default value if the symbolic
    # constant doesn't exist.
    default_webui_port = getattr(aiomonitor, 'MONITOR_WEBUI_PORT', 20102)
    parser.add_argument(
        '--aiomonitor', action='store_true', default=False,
        help='run aiomonitor debugging server')
    parser.add_argument(
        '--aiomonitor-host', type=str, default=aiomonitor.MONITOR_HOST,
        help='bind host for aiomonitor/aioconsole [%(default)s]')
    parser.add_argument(
        '--aiomonitor-port', type=int, default=aiomonitor.MONITOR_PORT,
        help='port for aiomonitor terminal UI [%(default)s]')
    parser.add_argument(
        '--aioconsole-port', type=int, default=aiomonitor.CONSOLE_PORT,
        help='port for aioconsole [%(default)s]')
    parser.add_argument(
        '--aiomonitor-webui-port', type=int, default=default_webui_port,
        help='port for aiomonitor web UI [%(default)s]')



################################################################################
# Copyright (c) 2017-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Simple utility functions to access information about network interfaces.

This is a simple wrapper around :mod:`netifaces`.
"""

import netifaces


def get_interface_address(interface):
    """Obtain the IPv4 address of a network interface.

    If the interface has multiple IPv4 addresses, it returns the first one.
    As a convenience, it is legal to pass ``None`` as the address, in which
    case ``None`` is returned.

    Parameters
    ----------
    interface : str
        Name of the network interface

    Returns
    -------
    address : str
        Dotted-quad representation of the IPv4 address

    Raises
    ------
    ValueError
        if the interface does not exist or does not have an IPv4 address
    """
    if interface is None:
        return None
    else:
        try:
            return netifaces.ifaddresses(interface)[netifaces.AF_INET][0]['addr']
        except (IndexError, KeyError):
            raise ValueError('No IPv4 address found for interface ' + interface)



################################################################################
# Copyright (c) 2017-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpservices.logging`"""

import io
import json
import logging
import os
import re
import signal
import socket
import sys
import time
import unittest
import zlib
from contextlib import closing
from unittest import mock

import katsdpservices


class TestLogging(unittest.TestCase):
    def _create_patch(self, *args, **kwargs):
        patcher = mock.patch(*args, **kwargs)
        mock_obj = patcher.start()
        self.addCleanup(patcher.stop)
        return mock_obj

    def _create_patch_dict(self, *args, **kwargs):
        patcher = mock.patch.dict(*args, **kwargs)
        mock_obj = patcher.start()
        self.addCleanup(patcher.stop)
        return mock_obj

    def _wipe_dummy_logger(self):
        logger = logging.getLogger('katsdpservices.test.dummy')
        for handler in list(logger.handlers):
            logger.removeHandler(handler)
        for filter in list(logger.filters):
            logger.removeFilter(filter)

    def setUp(self):
        # Grab the stderr written by the logger
        self.stderr = self._create_patch('sys.stderr', new_callable=io.StringIO)
        # Point root away from the actual root logger, so that we don't break
        # that.
        self.logger = self._create_patch(
            'logging.root', logging.getLogger('katsdpservices.test.dummy'))
        self.addCleanup(self._wipe_dummy_logger)
        # Wipe out existing environment, so that we can poke KATSDP_LOG_*
        self.environ = self._create_patch_dict(os.environ, clear=True)
        # Override time so that we can compare against a known value
        self.time = self._create_patch('time.time', autospec=True)
        self.time.return_value = 1488463323.125125
        self.addCleanup(signal.signal, signal.SIGHUP, signal.SIG_DFL)

    def test_simple(self):
        katsdpservices.setup_logging()
        logging.debug('debug message')
        logging.info('info message')
        logging.warning('warning message\nwith\nnewlines')
        self.assertRegex(
            self.stderr.getvalue(),
            re.compile(
                "\\A2017-03-02T14:02:03.125Z - test_logging.py:\\d+ - INFO - info message\n"
                "2017-03-02T14:02:03.125Z - test_logging.py:\\d+ - WARNING - warning message\n"
                "with\n"
                "newlines\n\\Z", re.M))

    def test_one_line(self):
        os.environ['KATSDP_LOG_ONELINE'] = '1'
        katsdpservices.setup_logging()
        logging.debug('debug message')
        logging.info('info message')
        logging.warning('warning message\nwith\nnewlines')
        # The \\\\ in the regex is a literal \: one level of escaping for
        # Python, one for regex.
        self.assertRegex(
            self.stderr.getvalue(),
            re.compile(
                "\\A2017-03-02T14:02:03.125Z - test_logging.py:\\d+ - INFO - info message\n"
                "2017-03-02T14:02:03.125Z - test_logging.py:\\d+ - WARNING - warning message\\\\ "
                "with\\\\ "
                "newlines\n\\Z", re.M))

    def test_log_level(self):
        os.environ['KATSDP_LOG_LEVEL'] = 'debug'
        katsdpservices.setup_logging()
        logging.debug('debug message')
        logging.info('info message')
        logging.warning('warning message\nwith\nnewlines')
        self.assertRegex(
            self.stderr.getvalue(),
            re.compile(
                "\\A2017-03-02T14:02:03.125Z - test_logging.py:\\d+ - DEBUG - debug message\n"
                "2017-03-02T14:02:03.125Z - test_logging.py:\\d+ - INFO - info message\n"
                "2017-03-02T14:02:03.125Z - test_logging.py:\\d+ - WARNING - warning message\n"
                "with\n"
                "newlines\n\\Z", re.M))

    def _test_gelf(self, localname, extra):
        self.maxDiff = 4096   # Diff tends to be too big to display otherwise
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        with closing(sock):
            sock.bind(('127.0.0.1', 0))
            port = sock.getsockname()[1]
            os.environ['KATSDP_LOG_GELF_ADDRESS'] = '127.0.0.1:{}'.format(port)
            if localname:
                os.environ['KATSDP_LOG_GELF_LOCALNAME'] = 'myhost'
            if extra:
                os.environ['KATSDP_LOG_GELF_EXTRA'] = '{"hello": "world", "number": 3}'
            # Fake the container ID
            container_id = "abcdef0123456789"
            with mock.patch('katsdpservices.logging.docker_container_id',
                            return_value=container_id):
                katsdpservices.setup_logging()
            # exc_info=False is to test the fix for
            # https://github.com/keeprocking/pygelf/issues/29
            logging.info('info message', exc_info=False)
            raw = sock.recv(4096)
            raw = zlib.decompress(raw)
        data = json.loads(raw.decode('utf-8'))
        # This dictionary may need to be updated depending on the implementation
        expected = {
            "timestamp": self.time.return_value,
            "version": "1.1",
            "short_message": "info message",
            "_logger_name": "katsdpservices.test.dummy",
            "_file": mock.ANY,
            "_line": mock.ANY,
            "_func": "_test_gelf",
            "_module": "test_logging",
            "_docker.id": container_id,
            "_timestamp_precise": "2017-03-02T14:02:03.125125Z",
            "level": 6,
            "host": "myhost" if localname else mock.ANY
        }
        if extra:
            expected["_hello"] = "world"
            expected["_number"] = 3
        expected["_stack_info"] = None
        if sys.version_info >= (3, 12, 0):
            expected["_taskName"] = None
        self.assertEqual(data, expected)

    def test_gelf_basic(self):
        self._test_gelf(False, False)

    def test_gelf_options(self):
        self._test_gelf(True, True)

    def test_toggle_debug(self):
        self.assertEqual(logging.INFO, logging.root.level)
        os.kill(os.getpid(), signal.SIGUSR2)
        # Give it a bit of time, since it's done in a separate thread
        time.sleep(0.01)
        self.assertEqual(logging.DEBUG, logging.root.level)
        # Check that it toggles back again
        os.kill(os.getpid(), signal.SIGUSR2)
        time.sleep(0.01)
        self.assertEqual(logging.INFO, logging.root.level)



################################################################################
# Copyright (c) 2017-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for :mod:`katsdpservices.argparse`."""

import unittest
from unittest import mock

from katsdptelstate.endpoint import Endpoint

from katsdpservices import ArgumentParser


class MockException(Exception):
    """Exception class used for monkey-patching functions that don't return."""
    pass


class TestArgumentParser(unittest.TestCase):
    def _stub_get(self, name, default=None):
        return self.data.get(name, default)

    def setUp(self):
        # Set up a mock version of TelescopeState which applies for the whole test
        patcher = mock.patch('katsdptelstate.TelescopeState', autospec=True)
        self.addCleanup(patcher.stop)
        self.TelescopeState = patcher.start()
        self.TelescopeState.return_value.get = mock.MagicMock(side_effect=self._stub_get)
        # Create a fixture
        self.parser = ArgumentParser()
        self.parser.add_argument('positional', type=str)
        self.parser.add_argument('--int-arg', type=int, default=5)
        self.parser.add_argument('--float-arg', type=float, default=3.5)
        self.parser.add_argument('--no-default', type=str)
        self.parser.add_argument('--bool-arg', action='store_true', default=False)
        group = self.parser.add_argument_group('Group options')
        group.add_argument('--group-int', type=int, default=6)
        mutual = self.parser.add_mutually_exclusive_group()
        mutual.add_argument('--mutual-x', default='x')
        mutual.add_argument('--mutual-y', default='y')
        self.data = {
            'config': {
                'int_arg': 10,
                'float_arg': 4.5,
                'no_default': 'telstate',
                'bool_arg': True,
                'not_arg': 'should not be seen',
                'telstate': 'example.org',
                'group_int': 11,
                'mutual_y': 'z',
                'level1': {
                    'int_arg': 11,
                    'level2': {
                        'float_arg': 5.5
                    }
                }
            },
            'config.level1.level2': {
                'float_arg': 12.5
            }
        }

    def test_no_telstate(self):
        """Passing explicit arguments but no telescope model sets the arguments."""
        args = self.parser.parse_args(
            ['hello', '--int-arg=3', '--float-arg=2.5', '--no-default=test', '--bool-arg',
             '--group-int=11', '--mutual-y=z'])
        self.assertIsNone(args.telstate)
        self.assertIsNone(args.telstate_endpoint)
        self.assertEqual('', args.name)
        self.assertEqual('hello', args.positional)
        self.assertEqual(3, args.int_arg)
        self.assertEqual(2.5, args.float_arg)
        self.assertEqual('test', args.no_default)
        self.assertEqual(True, args.bool_arg)
        self.assertEqual(11, args.group_int)
        self.assertEqual('z', args.mutual_y)

    def test_no_telstate_defaults(self):
        """Passing no optional arguments sets those arguments to default"""
        args = self.parser.parse_args(['hello'])
        self.assertIsNone(args.telstate)
        self.assertEqual('', args.name)
        self.assertEqual('hello', args.positional)
        self.assertEqual(5, args.int_arg)
        self.assertEqual(3.5, args.float_arg)
        self.assertIsNone(args.no_default)
        self.assertEqual(False, args.bool_arg)
        self.assertEqual(6, args.group_int)
        self.assertEqual('y', args.mutual_y)

    def test_telstate_no_name(self):
        """Passing --telstate but not --name loads from root config"""
        args = self.parser.parse_args(['hello', '--telstate=example.com'])
        self.assertIs(self.TelescopeState.return_value, args.telstate)
        self.assertEqual(Endpoint('example.com', 6379), args.telstate_endpoint)
        self.assertEqual('', args.name)
        self.assertEqual(10, args.int_arg)
        self.assertEqual(4.5, args.float_arg)
        self.assertEqual('telstate', args.no_default)
        self.assertEqual(True, args.bool_arg)
        self.assertEqual(11, args.group_int)
        self.assertEqual('z', args.mutual_y)
        self.assertNotIn('help', vars(args))
        self.assertNotIn('not_arg', vars(args))

    def test_telstate_nested(self):
        """Passing a nested name loads from all levels of the hierarchy"""
        args = self.parser.parse_args(['hello', '--telstate=example.com', '--name=level1.level2'])
        self.assertIs(self.TelescopeState.return_value, args.telstate)
        self.assertEqual(Endpoint('example.com', 6379), args.telstate_endpoint)
        self.assertEqual('level1.level2', args.name)
        self.assertEqual('hello', args.positional)
        self.assertEqual(11, args.int_arg)
        self.assertEqual(12.5, args.float_arg)
        self.assertEqual('telstate', args.no_default)

    def test_telstate_override(self):
        """Command-line parameters override telescope state"""
        args = self.parser.parse_args(
            ['hello', '--int-arg=0', '--float-arg=0', '--telstate=example.com',
             '--name=level1.level2', '--group-int=15'])
        self.assertIs(self.TelescopeState.return_value, args.telstate)
        self.assertEqual('level1.level2', args.name)
        self.assertEqual('hello', args.positional)
        self.assertEqual(0, args.int_arg)
        self.assertEqual(0.0, args.float_arg)
        self.assertEqual('telstate', args.no_default)
        self.assertEqual(15, args.group_int)

    def test_default_telstate(self):
        """Calling `set_default` with `telstate` keyword works"""
        self.parser.set_defaults(telstate='example.com')
        args = self.parser.parse_args(['hello'])
        self.TelescopeState.assert_called_once_with('example.com')
        self.assertIs(self.TelescopeState.return_value, args.telstate)
        self.assertEqual(10, args.int_arg)

    def test_convert_argument(self):
        """String argument in telescope state that must be converted to the appropriate type."""
        self.data['config']['int_arg'] = '50'
        args = self.parser.parse_args(['--telstate=example.com', 'hello'])
        self.assertEqual(50, args.int_arg)

    def test_bad_argument(self):
        """String argument in telescope state that cannot be converted must raise an error."""
        self.data['config']['int_arg'] = 'not an int'
        # We make the mock raise an exception, since the patched code is not
        # expecting the function to return.
        with mock.patch.object(
                self.parser, 'error', autospec=True, side_effect=MockException) as mock_error:
            with self.assertRaises(MockException):
                self.parser.parse_args(['--telstate=example.com', 'hello'])
            mock_error.assert_called_once_with(mock.ANY)

    def test_help(self):
        """Passing --help prints help without trying to construct the telescope state"""
        # We make the mock raise an exception, since the patched code is not
        # expecting the function to return.
        with mock.patch.object(
                self.parser, 'exit', autospec=True, side_effect=MockException) as mock_exit:
            with self.assertRaises(MockException):
                self.parser.parse_args(['--telstate=example.com', '--help'])
            mock_exit.assert_called_once_with()
            # Make sure we did not try to construct a telescope state
            self.assertEqual([], self.TelescopeState.call_args_list)



################################################################################
# Copyright (c) 2017-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Utility to make the process restart itself on SIGHUP"""

import sys
import os.path
import fcntl
import signal
import logging
import threading


# Latch the absolute path now, in case something changes directory later
# and makes a relative path invalid.
_restart_args = list(sys.argv)
_restart_args[0] = os.path.abspath(sys.argv[0])
_logger = logging.getLogger(__name__)


def restart_process():
    """Re-exec the process with its original arguments"""
    # Set file handles to close on exec, so that sockets don't live on
    # and prevent us opening the ports again.
    try:
        for name in os.listdir('/proc/self/fd'):
            try:
                fd = int(name)
            except ValueError:
                pass
            else:
                if fd > 2:   # Don't close stdin/stdout/stderr
                    try:
                        flags = fcntl.fcntl(fd, fcntl.F_GETFD)
                        fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
                    except OSError:
                        # This is guaranteed to happen, because the fd used by
                        # os.listdir will be in the list, but is closed by the
                        # time we try to change the flags on it.
                        pass
    except OSError:
        logging.warn('Could not read /proc/self/fd')
    # Ensure any logging gets properly flushed
    sys.stdout.flush()
    sys.stderr.flush()
    os.execlp(sys.executable, sys.executable, *_restart_args)


def setup_restart(signum=signal.SIGHUP, restart_callback=None):
    """Install a signal handler that calls :func:`restart` when :const:`SIGHUP`
    is received.

    Parameters
    ----------
    signum : int, optional
        Signum number for the signal handler to install
    restart_callback : callable
        If specified, this is called before execing. This can be used
        to cleanly wind up state. If it returns a true value, the restart is
        not done. This can be used if the callback just schedules cleanup and
        :func:`restart_process` in an asynchronous manner.
    """
    def restart_thread():
        logger = logging.getLogger(__file__)
        logger.warning("Received signal %d, restarting", signum)
        if restart_callback is not None:
            if restart_callback():
                return
        restart_process()

    def restart_handler(signum, frame):
        # It's not safe to log directly from a signal handler, so start
        # a separate thread to do the shutdown.
        thread = threading.Thread(target=restart_thread)
        thread.daemon = True
        thread.start()
    signal.signal(signal.SIGHUP, restart_handler)



################################################################################
# Copyright (c) 2017-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

r"""Customised SDP logging

Logging is controlled by environment variables:

KATSDP_LOG_ONELINE: if set (to any value), newlines in log messages are escaped
  to fit the message onto a single line (see :class:`OnelineFormatter`).
KATSDP_LOG_LEVEL: if set, it is used as the name of the log level. Otherwise,
  the log level defaults to INFO.
KATSDP_LOG_GELF_ADDRESS: if set (to a host:port), logging is sent over UDP
  to this address in Graylog Extended Logging Format.
KATSDP_LOG_GELF_LOCALNAME: if set, this overrides the local system name used
  in GELF log messages
KATSDP_LOG_GELF_EXTRA: set to a JSON dictionary (containing only strings and
  numbers, and with keys matching ``^[\w\.\-]*$``) of extra values to pass in
  every log message.

A signal handler is installed that toggles debug-level logging when SIGUSR2 is
received, and exception hooks are installed so that unhandled exceptions are
logged rather than merely printing to stderr.
"""

import logging
import os
import sys
import re
import time
import signal
import socket
import threading
import json
import datetime

import pygelf


_toggle_next_level = logging.DEBUG
"""Log level to set on next call to :func:`toggle_debug`."""


class OnelineFormatter(logging.Formatter):
    r"""Formatting that encodes newlines so that a single log message always
    appears on a single line. Newlines are replaced by "\ "  and backslashes
    are replaced by "\\".
    """
    def format(self, record):
        s = super().format(record)
        return s.replace('\\', r'\\').replace('\n', r'\ ')


def toggle_debug():
    """Swap current log level with the saved log level."""
    global _toggle_next_level
    old = logging.root.level
    logging.info(
        'Changing log level from %s to %s',
        logging.getLevelName(old), logging.getLevelName(_toggle_next_level))
    logging.root.setLevel(_toggle_next_level)
    _toggle_next_level = old


def _toggle_debug_handler(signum, frame):
    """Signal handler that calls :func:`toggle_debug` asynchronously.

    It's not safe to manipulate logging if the signal was received during a
    logging call. Instead, start a separate thread, and rely on logging's
    locking to do this change safely.
    """
    thread = threading.Thread(target=toggle_debug)
    thread.daemon = True
    thread.start()


def _setup_logging_stderr():
    if 'KATSDP_LOG_ONELINE' in os.environ:
        formatter_class = OnelineFormatter
    else:
        formatter_class = logging.Formatter
    formatter = formatter_class(
        "%(asctime)s.%(msecs)03dZ - %(filename)s:%(lineno)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S")
    formatter.converter = time.gmtime
    sh = logging.StreamHandler()
    sh.setFormatter(formatter)
    logging.root.addHandler(sh)


def docker_container_id():
    """Find the container ID of the current container.

    If not running in a container, returns ``None``.
    """
    regex = re.compile(':/docker/([a-f0-9]+)$')
    try:
        with open("/proc/self/cgroup") as f:
            for line in f:
                match = regex.search(line)
                if match:
                    return match.group(1)
    except OSError:
        pass
    return None


class _TimestampFilter(logging.Filter):
    """Adds an extra timestamp_precise field to records.

    The timestamp is formatted according to ISO8601 with microsecond
    precision. This is to work around the lack of native sub-ms
    timestamp support in logstash 7.x
    (https://github.com/elastic/logstash/issues/10822).

    While Elasticsearch supports nanosecond timestamps, LogRecord
    stores the creation time as a 64-bit float so is limited to about 200ns
    precision, and it doesn't seem worth the effort to go finer than
    microseconds.
    """
    def filter(self, record):
        dt = datetime.datetime.fromtimestamp(record.created, tz=datetime.timezone.utc)
        record.timestamp_precise = dt.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        return True


def _setup_logging_gelf():
    parts = os.environ['KATSDP_LOG_GELF_ADDRESS'].rsplit(':', 1)
    host = parts[0]
    if len(parts) == 2:
        port = parts[1]
    else:
        port = 12201     # Default GELF port

    # Resolve the hostname once up front, to work around
    # https://github.com/keeprocking/pygelf/issues/51. Restrict to IPv4
    # since it's not clear if pygelf can handle IPv6 (due to
    # https://bugs.python.org/issue14855).
    for res in socket.getaddrinfo(
            host, port,
            family=socket.AF_INET, type=socket.SOCK_DGRAM, proto=socket.IPPROTO_UDP):
        host, port = res[4][:2]
        break

    localname = os.environ.get('KATSDP_LOG_GELF_LOCALNAME')
    extras = os.environ.get('KATSDP_LOG_GELF_EXTRA', '{}')
    extras = json.loads(extras)
    if not isinstance(extras, dict):
        raise ValueError('KATSDP_LOG_GELF_EXTRA must be a JSON dict')
    # pygelf wants the additional fields to have _ already prepended
    extras = {'_' + key: value for (key, value) in extras.items()}
    container_id = docker_container_id()
    if container_id is not None:
        extras['_docker.id'] = container_id

    handler = pygelf.GelfUdpHandler(host, port,
                                    debug=True, include_extra_fields=True, compress=True,
                                    static_fields=extras)
    handler.addFilter(_TimestampFilter())
    if localname:
        handler.domain = localname
    logging.root.addHandler(handler)


def _sys_excepthook(exc_type, exc_value, exc_traceback):
    logging.error("Unhandled exception", exc_info=(exc_type, exc_value, exc_traceback))


def _threading_excepthook(args):
    if args.exc_type == SystemExit:
        return      # To match the behaviour of the default
    if logging is None or threading is None:
        return      # We're deep into interpreter shutdown - just give up

    # Based on Python 3.8 implementation: https://github.com/python/cpython/pull/13515
    if args.thread is not None:
        name = args.thread.name
    else:
        name = threading.get_ident()
    logging.error("Exception in thread {}:".format(name),
                  exc_info=(args.exc_type, args.exc_value, args.exc_traceback))


def setup_logging(add_signal_handler=True, add_excepthook=True):
    """Prepare logging. See the module-level documentation for details."""
    if os.environ.get('KATSDP_LOG_GELF_ADDRESS'):
        _setup_logging_gelf()
    _setup_logging_stderr()
    if 'KATSDP_LOG_LEVEL' in os.environ:
        logging.root.setLevel(os.environ['KATSDP_LOG_LEVEL'].upper())
    else:
        logging.root.setLevel(logging.INFO)
    logging.captureWarnings(True)
    if add_signal_handler:
        signal.signal(signal.SIGUSR2, lambda signum, frame: toggle_debug())
    if add_excepthook:
        sys.excepthook = _sys_excepthook
        # Only supported from Python 3.8
        if hasattr(threading, 'excepthook'):
            threading.excepthook = _threading_excepthook
